/*
 * copyright (c)2009 http://wamsoft.jp
 * zlib license
 */

#ifndef __SQOBJECTINFO_H__
#define __SQOBJECTINFO_H__

#include <stdio.h>
#include <squirrel.h>
#include <sqstdstring.h>
#include <string>

#ifdef USESQPLUS
#include <sqplus.h>
#else
#ifdef USESQRAT
#include <sqrat.h>
#else
#include "sqfunc.h"
#endif
#endif

#ifndef SQHEAPDEFINE
#define SQHEAPDEFINE \
	static void* operator new(size_t size) { return sq_malloc(size); }\
	static void* operator new[](size_t size) { return sq_malloc(size); }\
	static void  operator delete(void* p) { sq_free(p, 0); }\
	static void  operator delete[](void* p) { sq_free(p, 0); }
#endif

namespace sqobject {

class Object;

/// 
HSQUIRRELVM init();
/// I
extern void done();
/// ێpO[oVM̎擾
extern HSQUIRRELVM getGlobalVM();

// ---------------------------------------------------------
// StackValue
// ---------------------------------------------------------

/**
 * X^bNQƗp
 */
class StackValue {
	
public:
  // RXgN^
  StackValue(HSQUIRRELVM v, int idx) : v(v), idx(idx) {};
  
  // Cӌ^ւ̃LXg
  // 擾łȂꍇ̓NAlɂȂ
  template<typename T>
  operator T() const
  {
	T value;
	if (SQ_FAILED(getValue(v, &value, idx))) {
	  clearValue(&value);
	}
	return value;
  }

  // IuWFNgPUSH
  void push(HSQUIRRELVM v) const {
	  sq_move(v, this->v, idx);
  }

  // ^Ԃ
  SQObjectType type() const {
	return sq_gettype(v, idx);
  }

private:
  HSQUIRRELVM v;
  int idx;
};


// ---------------------------------------------------------
// ObjectInfo
// ---------------------------------------------------------

/**
 * squirrel IuWFNgێpNX
 * QƂێ\
 */
class ObjectInfo {
public:
#ifdef SQOBJHEAP
	SQHEAPDEFINE;
#endif
	// roottable ̎擾
	static ObjectInfo getRoot();
	// z̍쐬
	static ObjectInfo createArray(SQInteger size=0);
	// ̍쐬
	static ObjectInfo createTable();

	// e
	void clear();

	// X^bN擾
	void getStack(HSQUIRRELVM v, SQInteger idx);

    // X^bNQƂƂĎ擾
	void getStackWeak(HSQUIRRELVM v, SQInteger idx);
  
	// RXgN^
	ObjectInfo();

	// RXgN^
	ObjectInfo(HSQOBJECT obj);
	
	// RXgN^
	ObjectInfo(HSQUIRRELVM v, SQInteger idx);

	// Rs[RXgN^
	ObjectInfo(const ObjectInfo &orig);

	// 
	ObjectInfo & operator=(const ObjectInfo &orig);

	// Cӌ^̃RXgN^
	template<typename T>
	ObjectInfo(T value) {
		sq_resetobject(&obj);
		setValue(value);
	}

	// Cӌ^̑
	template<typename T>
	ObjectInfo & operator=(T value) {
		setValue(value);
		return *this;
	}

	// Cӌ^ւ̃LXg
	// 擾łȂꍇ̓NAlɂȂ
	template<typename T>
	operator T() const
	{
		HSQUIRRELVM gv = getGlobalVM();
		push(gv);
		T value;
		if (SQ_FAILED(getValue(gv, &value))) {
			clearValue(&value);
		}
		sq_pop(gv, 1);
		return value;
	}
	
	// l̐ݒ
	template<typename T>
	void setValue(T value) {
		HSQUIRRELVM gv = getGlobalVM();
		pushValue(gv, value);
		getStack(gv, -1);
		sq_pop(gv, 1);
	}
	
	// fXgN^
	virtual ~ObjectInfo();

	// null H
	bool isNull() const;
	
	// XbhH
	bool isSameThread(const HSQUIRRELVM v) const;

	// Xbh擾
	operator HSQUIRRELVM() const;
	
	// IuWFNgPUSH
	void push(HSQUIRRELVM v) const;

	/// o^(sNULL)
	void pushClone(HSQUIRRELVM v) const;
	
	// ---------------------------------------------------
	// r֐Q
	// ---------------------------------------------------

	template<typename T>
	bool operator ==(T &value) {
		HSQUIRRELVM v = getGlobalVM();
		push(v);
		pushValue(v, value);
		bool cmp = sq_cmp(v) == 0;
		sq_pop(v,2);
		return cmp;
	}

	template<typename T>
	bool operator !=(T &value) {
		HSQUIRRELVM v = getGlobalVM();
		push(v);
		pushValue(v, value);
		bool cmp = sq_cmp(v) != 0;
		sq_pop(v,2);
		return cmp;
	}
	
	template<typename T>
	bool operator <(T &value) {
		HSQUIRRELVM v = getGlobalVM();
		push(v);
		pushValue(v, value);
		bool cmp = sq_cmp(v) < 0;
		sq_pop(v,2);
		return cmp;
	}
	
	template<typename T>
	bool operator <=(T &value) {
		HSQUIRRELVM v = getGlobalVM();
		push(v);
		pushValue(v, value);
		bool cmp = sq_cmp(v) <= 0;
		sq_pop(v,2);
		return cmp;
	}

	template<typename T>
	bool operator >(T &value) {
		HSQUIRRELVM v = getGlobalVM();
		push(v);
		pushValue(v, value);
		bool cmp = sq_cmp(v) > 0;
		sq_pop(v,2);
		return cmp;
	}
	
	template<typename T>
	bool operator >=(T &value) {
		HSQUIRRELVM v = getGlobalVM();
		push(v);
		pushValue(v, value);
		bool cmp = sq_cmp(v) >= 0;
		sq_pop(v,2);
		return cmp;
	}

	// ---------------------------------------------------
	// delegate p
	// ---------------------------------------------------

	// delegate Ƃċ@\邩ǂ
	bool isDelegate() const;

	// bindenv 邩ǂ
	bool isBindDelegate() const;

	// ---------------------------------------------------
	// f[^擾
	// ---------------------------------------------------

	const SQChar *getString();
	
	// ---------------------------------------------------
	// zEp
	// ---------------------------------------------------

	/// zƂď
	void initArray(SQInteger size=0);
	
	/// @return zȂ true
	bool isArray() const { return sq_isarray(obj); }

	/// zƂď
	void initTable();

	/// @return zȂ true
	bool isTable() const { return sq_istable(obj); }
	
	/// zɒlǉ
	SQRESULT append(HSQUIRRELVM v, SQInteger idx);

	/// zɔzǉ
	SQRESULT appendArray(ObjectInfo &array);
	
	/// zɒlǉ
	template<typename T>
	SQRESULT append(T value) {
		HSQUIRRELVM gv = getGlobalVM();
		push(gv);
		pushValue(gv, value);
		SQRESULT ret = sq_arrayappend(gv, -2);
		sq_pop(gv,1);
		return ret;
	}

	/// zɒl}
	template<typename T>
	SQRESULT insert(SQInteger index, T value) {
		HSQUIRRELVM gv = getGlobalVM();
		push(gv);
		pushValue(gv, value);
		SQRESULT ret = sq_arrayinsert(gv, -2, index);
		sq_pop(gv,1);
		return ret;
	}

	/// z񂩂w肳ꂽCfbNX̒l폜
	SQRESULT remove(SQInteger index) {
		HSQUIRRELVM gv = getGlobalVM();
		push(gv);
		SQRESULT ret = sq_arrayremove(gv, -1, index);
		sq_pop(gv,1);
		return ret;
	}

	/// z񂩂w肳ꂽl폜
	template<typename T>
	SQRESULT removeValue(T value, bool all=false) {
		HSQUIRRELVM gv = getGlobalVM();
		push(gv);
		pushValue(gv, value);
		SQRESULT ret = sq_arrayremovevalue(gv, -2, all ? SQTrue : SQFalse);
		sq_pop(gv,1);
		return ret;
	}
	
	/// z/ɒli[
	template<typename K, typename T>
	SQRESULT set(K key, T value) {
		HSQUIRRELVM gv = getGlobalVM();
		push(gv);
		pushValue(gv, key);
		pushValue(gv, value);
		SQRESULT ret = sq_set(gv, -3);
		sq_pop(gv,1);
		return ret;
	}

	/// ɒlVKi[
	template<typename K, typename T>
	SQRESULT create(K key, T value) {
		HSQUIRRELVM gv = getGlobalVM();
		push(gv);
		pushValue(gv, key);
		pushValue(gv, value);
		SQRESULT ret = sq_newslot(gv, -3, SQFalse);
		sq_pop(gv,1);
		return ret;
	}

	/// l폜
	template<typename K>
	SQRESULT deleteslot(K key) {
		HSQUIRRELVM gv = getGlobalVM();
		push(gv);
		pushValue(gv, key);
		SQRESULT ret = sq_deleteslot(gv, -2, SQFalse);
		sq_pop(gv,1);
		return ret;
	}

	/// @return z/̒ key ݂Ȃ true Ԃ
	template<typename T>
	bool has(T key) const {
		HSQUIRRELVM gv = getGlobalVM();
		push(gv);
		pushValue(gv, key);
		if (sq_exists(gv, -2)) {
			sq_pop(gv, 1);
			return true;
		} else {
			sq_pop(gv, 1);
			return false;
		}
	}
	
	/// z/̒l擾B݂Ȃꍇ null Ԃ
	template<typename K>
	ObjectInfo get(K key) const {
		HSQUIRRELVM gv = getGlobalVM();
		pushData(gv, key);
		ObjectInfo ret(gv, -1);
		sq_pop(gv,1);
		return ret;
	}

	/**
	 * zQƗp
	 */
	class ObjectInfoReference {

	public:
		// RXgN^
		ObjectInfoReference(HSQOBJECT obj, int key) : obj(obj), intKey(key), type(0) {};
		ObjectInfoReference(HSQOBJECT obj, const SQChar *key) : obj(obj), strKey(key), type(1) {};
		ObjectInfoReference(HSQOBJECT obj, std::basic_string<SQChar> &key) : obj(obj), strKey(key.c_str()), type(1) {};
		
		// Cӌ^̑
		// ݒłȂꍇŎ̏ꍇ͍쐬Ă܂
		template<typename T>
		const T &operator=(const T &value) {
			ObjectInfo o(obj);
			if (o.type() == OT_ARRAY) {
				if (type == 0) {
					// TCY͂ĂȂ̂ null ł߂
					if (o.len() <= intKey) {
						HSQUIRRELVM gv = getGlobalVM();
						o.push(gv);
						while (o.len() <= intKey) {
							sq_pushnull(gv);
							sq_arrayappend(gv, -2);
						}
						sq_pop(gv,1);
					}
					o.set(intKey, value);
				}
			} else {
				SQRESULT ret;
				if (type == 0) {
					ret = o.set(intKey, value);
				} else {
					ret = o.set(strKey, value);
				}
				if (SQ_FAILED(ret)) {
					if (o.type() == OT_TABLE || o.type() == OT_CLASS) {
						if (type == 0) {
							o.create(intKey, value);
						} else {
							o.create(strKey, value);
						}
					}
				}
			}
			return value;
		}
		
		// Cӌ^ւ̃LXg
		// ϊłȂꍇ͏l(0)ł̃NAɂȂ
		template<typename T>
		operator T() const {
			ObjectInfo o(obj);
			T value;
			SQRESULT ret;
			if (type == 0) {
				ret = o.get(intKey, &value);
			} else {
				ret = o.get(strKey, &value);
			}
			if (SQ_FAILED(ret)) {
				clearValue(&value);
			}
			return value;
		}

		template<typename K>
		ObjectInfoReference operator[](K key) {
			HSQOBJECT target;
			HSQUIRRELVM gv = getGlobalVM();
			pushData(gv);
			sq_getstackobj(gv, -1, &target);
			ObjectInfoReference ret = ObjectInfoReference(target, key);
			sq_pop(gv, 1);
			return ret;
		}
		
		// f[^擾p
		void pushData(HSQUIRRELVM v) const {
			ObjectInfo o(obj);
			if (type == 0) {
				o.pushData(v, intKey);
			} else {
				o.pushData(v, strKey);
			}
		}
		
	protected:
		HSQOBJECT obj;
		int intKey;
		const SQChar *strKey;
		int type;
	};
	
	// l̐ݒ
	// QƏꍇp
	void setValue(ObjectInfoReference &ref) {
		HSQUIRRELVM gv = getGlobalVM();
		ref.pushData(gv);
		getStack(gv, -1);
		sq_pop(gv, 1);
	}
	
	/**
	 * z/QƗpt@X擾
	 * ANZXꍇ
	 * Es玩ō쐬
	 * E擾s珉l(0)Ԃ
	 * zANZXꍇ
	 * ETCY玩Iɂ̃TCY܂ő₷(null߂)
	 * E擾s珉l(0)Ԃ
	 * ƂIɍs܂
	 */
	template<typename K>
	ObjectInfoReference operator[](K key) {
		return ObjectInfoReference(obj, key);
	}

	/// z/̒l擾
	template<typename K, typename T>
	SQRESULT get(K key, T *value) const {
		SQRESULT ret;
		HSQUIRRELVM gv = getGlobalVM();
		push(gv);
		pushValue(gv, key);
		if (SQ_SUCCEEDED(ret = sq_get(gv, -2))) {
			ret = getValue(gv, value);
			sq_pop(gv,1);
		}
		sq_pop(gv,1);
		return ret;
	}
	
	/// z/̒lpushBȂnullpush
	template<typename T>
	void pushData(HSQUIRRELVM v, T key) const {
		push(v);
		pushValue(v, key);
		if (SQ_FAILED(sq_get(v, -2))) {
			sq_pushnull(v);
		}
		sq_remove(v,-2);
	}
	
	/// zE̒gNA
	void clearData() {
		HSQUIRRELVM gv = getGlobalVM();
		push(gv);
		sq_clear(gv,-1);
		sq_pop(gv,1);
	}

	/// @return IuWFNǧ^
	SQObjectType type() const {
		return obj._type;
	}
	
	/// @return z̒/̃TCY/̒
	SQInteger len() const;

	/**
	 * z̓eSPUSH
	 * @param v squirrelVM
	 * @return push 
	 */
	SQInteger pushArray(HSQUIRRELVM v) const;

	// ---------------------------------------------------
	// ֐p
	// ---------------------------------------------------

	/// @return zȂ true
	bool isClosure() const { return sq_isclosure(obj) || sq_isnativeclosure(obj); }

	// Ăяo
	SQRESULT call(ObjectInfo *self=NULL);

	template<typename T1> SQRESULT call(T1 p1, ObjectInfo *self=NULL) {
		HSQUIRRELVM gv = getGlobalVM();
		push(gv);
		if (self) {
			self->push(gv);
		} else {
			sq_pushroottable(gv); // root
		}
		pushValue(gv, p1);
		SQRESULT ret = sq_call(gv, 2, SQFalse, SQTrue);
		sq_pop(gv, 1);
		return ret;
	}

	template<typename T1, typename T2> SQRESULT call(T1 p1, T2 p2, ObjectInfo *self=NULL) {
		HSQUIRRELVM gv = getGlobalVM();
		push(gv);
		if (self) {
			self->push(gv);
		} else {
			sq_pushroottable(gv); // root
		}
		pushValue(gv, p1);
		pushValue(gv, p2);
		SQRESULT ret = sq_call(gv, 3, SQFalse, SQTrue);
		sq_pop(gv, 1);
		return ret;
	}
	
	template<typename R> SQRESULT callResult(R* r, ObjectInfo *self=NULL) {
		SQRESULT ret;
		HSQUIRRELVM gv = getGlobalVM();
		push(gv);
		if (self) {
			self->push(gv);
		} else {
			sq_pushroottable(gv); // root
		}
		if (SQ_SUCCEEDED(ret = sq_call(gv, 1, SQTrue, SQTrue))) {
			ret = getResultValue(gv, r);
			sq_pop(gv, 1);
		}
		sq_pop(gv, 1);
		return ret;
	}
	template<typename R, typename T1> SQRESULT callResult(R* r, T1 p1, ObjectInfo *self=NULL) {
		SQRESULT ret;
		HSQUIRRELVM gv = getGlobalVM();
		push(gv);
		if (self) {
			self->push(gv);
		} else {
			sq_pushroottable(gv); // root
		}
		pushValue(gv, p1);
		if (SQ_SUCCEEDED(ret = sq_call(gv, 2, SQTrue, SQTrue))) {
			ret = getResultValue(gv, r);
			sq_pop(gv, 1);
		}
		sq_pop(gv, 1);
		return ret;
	}
	template<typename R, typename T1, typename T2> SQRESULT callResult(R* r, T1 p1, T2 p2, ObjectInfo *self=NULL) {
		SQRESULT ret;
		HSQUIRRELVM gv = getGlobalVM();
		push(gv);
		if (self) {
			self->push(gv);
		} else {
			sq_pushroottable(gv); // root
		}
		pushValue(gv, p1);
		pushValue(gv, p2);
		if (SQ_SUCCEEDED(ret = sq_call(gv, 3, SQTrue, SQTrue))) {
			ret = getResultValue(gv, r);
			sq_pop(gv, 1);
		}
		sq_pop(gv, 1);
		return ret;
	}

	/**
	 * ȃIuWFNg\bhĂяoi)
	 * @param methodName \bh
	 */
	SQRESULT callMethod(const SQChar *methodName) {
		if (!isNull()) {
			ObjectInfo method = get(methodName);
			if (method.isClosure()) {
				return method.call(this);
			}
		}
		return SQ_ERROR;
	}

	/**
	 * ȃIuWFNg\bhĂяoi1)
	 * @param methodName \bh
	 * @param p1 
	 */
	template<typename T1> SQRESULT callMethod(const SQChar *methodName, T1 p1) {
		if (!isNull()) {
			ObjectInfo method = get(methodName);
			if (method.isClosure()) {
				return method.call(p1, this);
			}
		}
		return SQ_ERROR;
	}
	
	/**
	 * ȃIuWFNg\bhĂяoi2)
	 * @param methodName \bh
	 * @param p1 
	 * @param p2 2
	 */
	template<typename T1, typename T2> SQRESULT callMethod(const SQChar *methodName, T1 p1, T2 p2) {
		if (!isNull()) {
			ObjectInfo method = get(methodName);
			if (method.isClosure()) {
				return method.call(p1, p2, this);
			}
		}
		return SQ_ERROR;
	}
	
	/**
	 * ԒlL莩ȃIuWFNg\bhĂяoi)
	 * @param r Al|C^
	 * @param methodName \bh
	 */
	template<typename R> SQRESULT callMethodResult(R* r, const SQChar *methodName) {
		if (!isNull()) {
			ObjectInfo method = get(methodName);
			if (method.isClosure()) {
				return method.callResult(r, this);
			}
		}
		return SQ_ERROR;
	}

	/**
	 * Ԓl莩ȃIuWFNg\bhĂяoi1)
	 * @param r Al|C^
	 * @param methodName \bh
	 * @param p1 
	 */
	template<typename R, typename T1> SQRESULT callMethodResult(R* r, const SQChar *methodName, T1 p1) {
		if (!isNull()) {
			ObjectInfo method = get(methodName);
			if (method.isClosure()) {
				return method.callResult(r, p1, this);
			}
		}
		return SQ_ERROR;
	}
	
	/**
	 * ԒlL莩ȃIuWFNg\bhĂяoi2)
	 * @param r Al|C^
	 * @param methodName \bh
	 * @param p1 
	 * @param p2 2
	 */
	template<typename R, typename T1, typename T2> SQRESULT callMethodResult(R* r, const SQChar *methodName, T1 p1, T2 p2) {
		if (!isNull()) {
			ObjectInfo method = get(methodName);
			if (method.isClosure()) {
				return method.callResult(r, p1, p2, this);
			}
		}
		return SQ_ERROR;
	}

	// ---------------------------------------------------
	// NXp
	// ---------------------------------------------------

	// NXIuWFNg
	bool isClass() const;


	// ---------------------------------------------------
	// Ce[^
	// ---------------------------------------------------

	/**
	 * Ce[^Ăяo֐
	 * @param key L[
	 * @param value l
	 * @param userData [Uf[^
	 */
	typedef void (*foreachFunc)(const ObjectInfo &key, const ObjectInfo &value, void *userData);

	/**
	 * ֐ŃCe[^
	 * @param func Ăяo֐
	 * @param userData [Uf[^
	 */
	void foreach(foreachFunc func, void *userData=NULL) {
		HSQUIRRELVM gv = getGlobalVM();
		push(gv);
		sq_pushnull(gv);
		while (SQ_SUCCEEDED(sq_next(gv,-2))) {
			func(ObjectInfo(gv,-2), ObjectInfo(gv,-1), userData);
			sq_pop(gv,2);
		}
		sq_pop(gv,1); // null
	}

	/**
	 * ֐IuWFNgŃCe[^Bȉ֐IuWFNgn
	 * anytype  ObjectInfo LXg\Ȍ^ȂȂłOK
	 * struct Func {
	 *  void operator()(anytype key, anytype value, void *userData) {
	 * }
	 */
	template<class F>
	void foreach(F &func) {
		HSQUIRRELVM gv = getGlobalVM();
		push(gv);
		sq_pushnull(gv);
		while (SQ_SUCCEEDED(sq_next(gv,-2))) {
			func(ObjectInfo(gv,-2), ObjectInfo(gv, -1));
			sq_pop(gv,2);
		}
		sq_pop(gv,1); // null
	}

	// ---------------------------------------------------
	// 񏈗
	// ---------------------------------------------------
	
	/**
	 * \LԂ
	 * tostring() ̏B
	 */
	std::basic_string<SQChar> toString() const;

private:
	HSQOBJECT obj; // IuWFNgQƏ
};

// ---------------------------------------------------
// X^bNp
// ---------------------------------------------------

// Object push
// @return ł squirrel pɏς݂ŃCX^XĂ push ł true
bool pushObject(HSQUIRRELVM v, Object *obj);

// l push
void pushValue(HSQUIRRELVM v, bool value);
void pushValue(HSQUIRRELVM v, SQBool value);
void pushValue(HSQUIRRELVM v, SQInteger value);
void pushValue(HSQUIRRELVM v, SQFloat value);
void pushValue(HSQUIRRELVM v, const SQChar *value);
void pushValue(HSQUIRRELVM v, SQUserPointer value);
void pushValue(HSQUIRRELVM v, ObjectInfo &obj);
void pushValue(HSQUIRRELVM v, StackValue &sv);
void pushValue(HSQUIRRELVM v, std::basic_string<SQChar> &value);
void pushValue(HSQUIRRELVM v, SQFUNCTION func);
void pushValue(HSQUIRRELVM v, HSQOBJECT obj);

// l̎擾
SQRESULT getValue(HSQUIRRELVM v, bool *value, int idx=-1);
SQRESULT getValue(HSQUIRRELVM v, SQBool *value, int idx=-1);
SQRESULT getValue(HSQUIRRELVM v, SQInteger *value, int idx=-1);
SQRESULT getValue(HSQUIRRELVM v, SQFloat *value, int idx=-1);
SQRESULT getValue(HSQUIRRELVM v, const SQChar **value, int idx=-1);
SQRESULT getValue(HSQUIRRELVM v, SQUserPointer *value, int idx=-1);
SQRESULT getValue(HSQUIRRELVM v, ObjectInfo *value, int idx=-1);
SQRESULT getValue(HSQUIRRELVM v, std::basic_string<SQChar> *value, int idx=-1);

#ifdef USESQPLUS
// l̊i[
// i[sƂ̓IuWFNg폜̂ delete ̕Kv͂Ȃ
//  squirrel ŐꂽIuWFNgꍇ͕K
template<typename T>
void pushValue(HSQUIRRELVM v, T *value) {
	if (value) {
		if (pushObject(v, value) || SQ_SUCCEEDED(pushnewsqobj(v, GetTypeName(*value), value))) {
			return;
		}
	}
	sq_pushnull(v);
}

// l̊i[̑p
template<typename T>
void pushOtherValue(HSQUIRRELVM v, T *value) {
	if (value) {
		if (SQ_SUCCEEDED(pushotherobj(v, GetTypeName(*value), value))) {
			return;
		} else {
			delete value;
		}
	}
	sq_pushnull(v);
}

// l̎擾
template<typename T>
SQRESULT getValue(HSQUIRRELVM v, T **value, int idx=-1) {
	*value = SqPlus::GetInstance<T,false>(v, idx);
	return SQ_OK;
}
#else
#ifdef USESQRAT
// l̊i[
// i[sƂ̓IuWFNg폜̂ delete ̕Kv͂Ȃ
//  squirrel ŐꂽIuWFNgꍇ͕K
template<typename T>
void pushValue(HSQUIRRELVM v, T *value) {
	if (value) {
		if (pushObject(v, value)) {
			return;
		}
		Sqrat::Var<T*>::pushInit(v, value);
		return;
	}
	sq_pushnull(v);
}

// l̊i[̑p
template<typename T>
void pushOtherValue(HSQUIRRELVM v, T *value) {
	if (value) {
		Sqrat::Var<T*>::pushInit(v, value);
		return;
	}
	sq_pushnull(v);
}

// l̎擾
template<typename T>
SQRESULT getValue(HSQUIRRELVM v, T **value, int idx=-1) {
	*value = Sqrat::Var<T*>(v, idx).value;
	return SQ_OK;
}
#else

// l̊i[
// i[sƂ̓IuWFNg폜̂ delete ̕Kv͂Ȃ
//  squirrel ŐꂽIuWFNgꍇ͕K
template<typename T>
void pushValue(HSQUIRRELVM v, T *value) {
	if (value) {
		if (pushObject(v, value)) {
			SQClassType<T>::pushInstance(v, value);
			return;
		}
	}
	sq_pushnull(v);
}

// l̎擾
template<typename T>
SQRESULT getValue(HSQUIRRELVM v, T **value, int idx=-1) {
	*value = SQClassType<T>::getInstance(v, idx);
	return SQ_OK;
}
#endif
#endif

// l̋
void clearValue(bool *value);
void clearValue(SQBool *value);
void clearValue(SQInteger *value);
void clearValue(SQFloat *value);
void clearValue(const SQChar **value);
void clearValue(SQUserPointer *value);
void clearValue(ObjectInfo *value);
void clearValue(std::basic_string<SQChar> *value);

// l̋
template<typename T>
void clearValue(T **value) {
	*value = 0;
}

// l̎擾F{ getValue ̃RsyB͈SłȂꍇ̂ŔrKv
SQRESULT getResultValue(HSQUIRRELVM v, bool *value);
SQRESULT getResultValue(HSQUIRRELVM v, SQBool *value);
SQRESULT getResultValue(HSQUIRRELVM v, SQInteger *value);
SQRESULT getResultValue(HSQUIRRELVM v, SQFloat *value);
SQRESULT getResultValue(HSQUIRRELVM v, SQUserPointer *value);
SQRESULT getResultValue(HSQUIRRELVM v, ObjectInfo *value);
SQRESULT getResultValue(HSQUIRRELVM v, std::basic_string<SQChar> *value);

// l̎擾
template<typename T>
SQRESULT getResultValue(HSQUIRRELVM v, T **value) {
	return getValue(value);
}

// --------------------------------------------------------------------------------------
// printf
// --------------------------------------------------------------------------------------

/**
 * printf̏
 * @param format 
 * @return \
 */
template<typename DUMMY>
inline SQInteger printf(const SQChar *format)
{
	HSQUIRRELVM gv = getGlobalVM();
	sq_pushstring(gv, format, -1);
	return sqstd_printf(gv, 0);
}

template<typename T1>
inline SQInteger printf(const SQChar *format, T1 p1) {
	HSQUIRRELVM gv = getGlobalVM();
	sq_pushstring(gv, format, -1);
	pushValue(gv, p1);
	return sqstd_printf(gv, 1);
}

template<typename T1, typename T2>
inline SQInteger printf(const SQChar *format, T1 p1, T2 p2) {
	HSQUIRRELVM gv = getGlobalVM();
	sq_pushstring(gv, format, -1);
	pushValue(gv, p1);
	pushValue(gv, p2);
	return sqstd_printf(gv, 2);
}

template<typename T1, typename T2, typename T3>
inline SQInteger printf(const SQChar *format, T1 p1, T2 p2, T3 p3) {
	HSQUIRRELVM gv = getGlobalVM();
	sq_pushstring(gv, format, -1);
	pushValue(gv, p1);
	pushValue(gv, p2);
	pushValue(gv, p3);
	return sqstd_printf(gv, 3);
}

template<typename T1, typename T2, typename T3, typename T4>
inline SQInteger printf(const SQChar *format, T1 p1, T2 p2, T3 p3, T4 p4) {
	HSQUIRRELVM gv = getGlobalVM();
	sq_pushstring(gv, format, -1);
	pushValue(gv, p1);
	pushValue(gv, p2);
	pushValue(gv, p3);
	pushValue(gv, p4);
	return sqstd_printf(gv, 4);
}

template<typename T1, typename T2, typename T3, typename T4, typename T5>
inline SQInteger printf(const SQChar *format, T1 p1, T2 p2, T3 p3, T4 p4, T5 p5) {
	HSQUIRRELVM gv = getGlobalVM();
	sq_pushstring(gv, format, -1);
	pushValue(gv, p1);
	pushValue(gv, p2);
	pushValue(gv, p3);
	pushValue(gv, p4);
	pushValue(gv, p5);
	return sqstd_printf(gv, 5);
}

// --------------------------------------------------------------------------------------

/**
 * p
 */
class StackInfo {

public:
	/**
	 * RXgN^
	 * @param vm VM
	 */
	StackInfo(HSQUIRRELVM vm) : vm(vm) {
		argc = sq_gettop(vm) - 1;
	}
	
	/**
	 * @return ̐
	 */
	int len() const {
		return argc;
	}

	/**
	 * @return selfQ
	 */
	ObjectInfo getSelf() const {
		return ObjectInfo(vm, 1);
	}

	/**
	 * @param n ԍ 0`
	 * @return ̌^
	 */
	SQObjectType getType(int n) const {
		if (n < argc) {
			return sq_gettype(vm, n+2);
		}
		return OT_NULL;
	}
	
	/**
	 * @param n ԍ 0`
	 * @return ̌^
	 */
	ObjectInfo getArg(int n) const {
		ObjectInfo ret;
		if (n < argc) {
			ret.getStack(vm, n+2);
		}
		return ret;
	}

	/**
	 * 擾
	 * @param n ԍ 0`
	 * @return IuWFNg
	 */
	ObjectInfo operator[](int n) const {
		return getArg(n);
	}

	// ʓo^
	SQRESULT setReturn() const { return 0; }
	template<typename T>
	SQRESULT setReturn(T value) const {
		pushValue(vm, value);
		return 1;
	}

protected:
	HSQUIRRELVM vm; //< VM
	SQInteger argc; //< ̐
};

};

#endif
